import os, re
from pathlib import Path
import subprocess
from typing import Callable, Optional
from rich import inspect, print
from functools import reduce
from more_itertools import peekable
import operator

flatten = lambda l: reduce(operator.iconcat, l, [])
p = lambda var: print(repr(var))
split_keep = lambda s, d: [e+d for e in s.split(d) if e]

source = (Path(__file__).parent / "src.itex").read_text()

unicode_map = {
    "∞": r"\infty",
    "≥": r"\ge",
    "≤": r"\le",
    "|->": r"\mapsto",
    "$": r"\textdollar",
    "~=": r"\approx",
    "¹": r"^1",
    "²": r"^2",
    "³": r"^3",
    "⁴": r"^4",
    "⁵": r"^5",
    "⁶": r"^6",
    "⁷": r"^7",
    "⁸": r"^8",
    "⁹": r"^9",
    "⁰": r"^0",
    "Ø": r"\emptyset",
    "€": r"\in",
    "·": r"\cdot",
    "×": r"\times",
    "%": r"\textpercent",
    "->": r"\to",
    "<-": r"\from",
    "=>": r"\implies",
    "<>": r"\diamond",
}


class Macro:
    source: str
    args_count: int

    def __init__(self, source: str):
        self.source = source
        self.args_count = len(re.findall(r"#\d+", source))

    def __call__(self, *args: str):
        if len(args) != self.args_count:
            raise TypeError(
                f"This macro expected {self.args_count} arguments but received {len(args)}:\n"
                + "\n\t·".join(map(repr, args))
            )
        output = self.source
        for i, arg in enumerate(args):
            output = output.replace(f"#{i+1}", arg)
        return output


def split_preamble_document(source: str) -> tuple[str, str]:
    seen_separator = False
    document, preamble = str(), str()
    for line in source.splitlines():
        dashes_count = len([c for c in line.strip() if c == "-"])
        if dashes_count >= 3 and dashes_count == len(line.strip()):
            seen_separator = True
            continue
        if seen_separator:
            document += line + "\n"
        else:
            preamble += line + "\n"

    return preamble, document


def parse_preamble(source: str) -> tuple[str, dict[str, Macro], str]:
    """
    Returns:
    - generated LaTeX preamble source
    - dict mapping macro names to macro functions to call with the parameters)
    """
    macros: dict[str, Macro] = {}
    out = str()
    in_foreach = False
    in_if_block = False
    in_else_block = False
    in_decl_block = False
    current_branch_instructions = ""
    foreach_instructions = ""
    decl_block_name, decl_block_operator = "", ""
    decl_block_instructions = ""
    for line in source.splitlines():
        matches = lambda p: re.match(p, line.strip())
        if (match := matches(r"for each ([$#*]) in (.)\.{2}(.)")) :
            itervar, range_start, range_end = match.groups()
            in_foreach = True
        elif in_foreach and is_indented(line):
            foreach_instructions += line.strip() + "\n"
        elif in_foreach and not line.startswith("\t"):
            in_foreach = False
            for charcode in range(ord(range_start), ord(range_end) + 1):
                generated, macromap = parse_preamble(
                    foreach_instructions.replace(itervar, chr(charcode))
                )
                out += generated
                macros |= macromap
        elif (match := matches(r"if\s+(.+)")) :
            condition = match.groups()
            in_if_block = True
        elif in_if_block and is_indented(line):
            current_branch_instructions += line.strip() + "\n"
        elif in_if_block and line.strip() == "else":
            # TODO: generate
            in_else_block = True
            in_if_block = False
        elif in_if_block and not is_indented(line):
            in_if_block = False
            # TODO: generate
        elif in_else_block and is_indented(line):
            current_branch_instructions += line.strip() + "\n"
        elif in_else_block and not is_indented(line):
            in_else_block = False
            # TODO: generate
        elif (match := matches(r"(\w+) ([^\[=:]+)(?: (\[[^\]]+\]))?")) :
            command, content, options = match.groups()
            content = eval_shell_interpolation(content.strip())
            if command == "use":
                command = "usepackage"
            if command == "operators":
                for operatorname in content.split(","):
                    operatorname = operatorname.strip()
                    out += (r"\newcommand{\%s}{\operatorname{%s}}" + "\n") % (
                        operatorname,
                        operatorname,
                    )
            else:
                out += f"\\{command}{{{content}}}"
            if options:
                out += f"[{options}]"
            out += "\n"
        elif (match := matches(r"(\w+)\s+(:?=)\s+$")) :
            in_decl_block = True
            decl_block_name, decl_block_operator = match.groups()
        elif in_decl_block and is_indented(line):
            decl_block_instructions += line.strip() + "\n"
        elif in_decl_block and not is_indented(line):
            in_decl_block = False
        elif (match := matches(r"(\w+)\s+(:?=)\s+(.+)")) :
            name, _, content = match.groups()
            content = eval_shell_interpolation(content.strip())
            macros[name] = Macro(content)
    return out, macros


def is_indented(line: str) -> bool:
    return line.startswith("\t") or line.startswith(" " * 4)


def eval_shell_interpolation(content: str) -> str:
    if content.startswith("``") and content.endswith("``"):
        return (
            subprocess.run(content.replace("``", ""), shell=True, capture_output=True)
            .stdout.decode("utf-8")
            .removesuffix("\n")
        )
    """
	for match in re.finditer(r'\b((op|bb|c|b|f|k)"([^"]+)")', content):
		matching_source, fontname, text = match.groups()
		content = content.replace(
			matching_source, f"\\{get_font_name(fontname)}{{{text}}}"
		)
	for search, replace in unicode_map.items():
		content = content.replace(search, replace)
	"""
    return content


def parse_section(line: str, lineno: int) -> Optional[dict]:
    if (match := re.match(r"={3,}\s+(.+)$", line)) :
        SECTION_LEVEL = 1
        return {
            "line": {"contents": line, "number": lineno},
            "raw": match.group(1),
            "number": ...,
            "level": SECTION_LEVEL,
            "parsed": ...,
        }
    return None


RELATIONS = (
    "propto",
    "=",
    ":=",
    "<=>",
    ">",
    "<",
    "<=",
    ">=",
    "<==",
    "=>",
    "==>",
    "->",
    "<-",
    # '|->', ... is this a relation?
    "!=",
    "in",
    "~",
    "~~",
    "~=",
    "equiv",
    "cong",
)

QUANTIFIERS = ["forall", "exists", "exists!"]

"@pure"


def list_to_regex(
    *choices: list[str],
    whitespace_around: bool = False,
    whitespace_left: bool = False,
    whitespace_right: bool = False,
) -> re.Pattern:
    surround_left = " " if whitespace_around or whitespace_left else ""
    surround_right = " " if whitespace_around or whitespace_right else ""

    return re.compile(
        surround_left
        + r"("
        + "|".join([f"(?:{re.escape(choice)})" for choice in choices])
        + r")"
        + surround_right
    )


def parse_math_block(lines: list[str], position: range) -> tuple[dict, str]:
    """
    Parses lines from a math block (expects indentation to be removed)
    Returns:
    - dict: reference name |-> {referenced line's content, referenced line's position}
    - str: the parsed math
    """

    if len(lines) > 1:
        for line in lines:
            if "&" not in line:
                line = insert_alignment_marker(line)
            line += r" \\"  # TODO add after parsing the math

    # Add environment
    if len(lines) == 1:
        lines = [r"\["] + lines + [r"\]"]
    else:
        lines = [r"begin(align*)"] + lines + [r"end(align*)"]

def insert_alignment_marker(line: str) -> str:
    """
    Given a math expression line, places an alignment character
    at the centermost relation symbol, ignoring relation symbols used to quantify an expression
    (eg ignore "∈" in ∀a ∈ ℝ) and relation symbols inside parentheses (as those are likely used to eg. index a summation)

    >>> insert_alignment_marker('forall a, b in RR, forall f in c"C"([a, b], <blue>CC</blue>), exists K in RR_+, |f| <= K')
    'forall a, b in RR, forall f in c"C"([a, b], <blue>CC</blue>), exists K in RR_+, |f| &<= K'

    (Notice the &<= instead of <=)
    """
    # TODO: configurable RELATION_CHARS
    # TODO: configurable QUANTIFIERS

    # Stores matches of possible relation symbols
    # One will be chosen out of those
    candidates: list[re.Match] = []
    for symbol in list_to_regex(RELATION_SYMBOLS, whitespace_around=True).finditer(
        line
    ):
        behind = line[: symbol.span()[0]]
        # Check for parentheses
        # If there are more opened ones that closed ones, we _are_ insides parentheses
        if behind.count("(") > behind.count(")"):
            # It's inside parentheses, discard this symbol.
            continue
        # Check for quantifier ownership
        # If there are more quantifiers than relations before this relation,
        # the current relation is owned by a quantifier.
        # This is not perfect, as some authors _might_ shove multiple relationships into a single quantifier,
        # but until then, this is it.
        """
		Other algorithm:
		if there are quantifiers ahead, then it's a quantifier (provided that the author writes quantifiers first)
		else, 🤔🤔🤔🤔
		"""
        relations_behind = list_to_regex(RELATION_SYMBOLS, whitespace_around=True).findall(' ' + behind)
        quantifiers_behind = list_to_regex(QUANTIFIERS, whitespace_around=True).findall(' '+ behind)
        if len(quantifiers_behind) > len(relations_behind):
            # If it's owned by a quantifier, discard this symbol.
            continue
        
        
        candidates.append(symbol)

    # Sort by position (not technically necessary)
    candidates.sort(key=lambda s: s.span()[0])

    # Choose the candidate
    chosen_relation = candidates[int(len(candidates) / 2)]

    # Align!
    return line[: chosen_relation.span()[0]+1] + "&" + line[chosen_relation.span()[0]+1 :]

OPERATORS = {
        "prefix": ('¬',),
        "infix": '+×·*^-'.split(),
        "suffix": ('!',),
        "function": "sin cos tan tanh cosh sinh sh ch sgn sg id exp ln log".split(' ')
}

# NOTE: this is starting to look like an EBNF thing, nice!
# Regexp for an atom.
ATOM = list_to_regex(
    r"[a-zA-Z]+|\d+", # Letters & numbers
    r"\(.+\)", r"\[.+\]", r"\{.+\}", # grouping
    r"\|\|.+\|\|", # norms
    r"\|.+\|", # absolute values
    r'[boctks]?"([^"])"', # text
    *RELATIONS, # relation characters are atomic
    *OPERATORS["prefix"], # operators are atomic
    *OPERATORS["infix"], 
    *OPERATORS["suffix"], 
    *OPERATORS["function"], 
)
# Constructs that keep an atom atomic. Regexps.
ATOMIC_CONSTRUCTS = list_to_regex(
    f"{ATOM}_{ATOM}", # subscript
    f"{ATOM}__{ATOM}", # underscript
    f"{ATOM}\\^{ATOM}", # superscript
    f"{ATOM}\\^{{2}}{ATOM}", # overscript
    f"{ATOM}->", # vector shorthand
    f"{ATOM}\\\\.{{1,2}}", # dot derivate shorthand
    f"{ATOM}'{{1,4}}" # derivative
)
FRACTION = f"{ATOM}/{ATOM}"
DERIVATIVE = f"d{ATOM}?/d{ATOM}"

def parse_overscript(expr: str) -> str:
    """
    Transforms overscript constructs A^^B into overset(A)(B)

    >>> 
    """
    matches = peekable(re.finditer(f'({ATOM})\\^{{2}}({ATOM})'))
    parsed = expr[:matches[0].start()]
    for match in matches:
        base, on_top = match.groups()
        parsed += expr.replace(match.string, f"overset({on_top})({base})")
        if matches.peek(None):
            parsed += expr[match.end():matches.peek().start()]
        else:
            parsed += expr[match.end():]
    return parsed

def parse_underscript(expr: str) -> str:
    """
    Transforms undescript constructs A__B into underset(A)(B)

    >>> parse_underscript('W__(A->B)(P->) = ±mg(z_B - z_A) "si  (Oz)  est vertical ascendant"')
    underset(W)(A->B)(P->) = ±mg(z_B - z_A) "si  (Oz)  est vertical ascendant"
    """
    matches = peekable(re.finditer(f'(' + list_to_regex(ATOM) + r")__(" + list_to_regex(ATOM) + r')'))
    parsed = expr[:matches[0].start()]
    for match in matches:
        base, under_base = match.groups()
        parsed += expr.replace(match.string, f"underset({under_base})({base})")
        if matches.peek(None):
            parsed += expr[match.end():matches.peek().start()]
        else:
            parsed += expr[match.end():]
    return parsed


def parse_vector(expr: str) -> str:
    """
    Transforms the shorthand vector notation into a standard command call

    >>> parse_vector('OM->.. = (r.. - r theta.^2)__(a_r) u_r-> + (2 r. theta. + r theta..)__(a_theta) u_theta-> + z.. u_z->')
    'vec(OM).. = (r.. - r theta.^2)__(a_r) vec(u_r) + (2 r. theta. + r theta..)__(a_theta) vec(u_theta) + z.. vec(u_z)'
    """
    matches = peekable(re.finditer(r"(^| +)((?:[^ ]+)|(?:\([^)]+\)))->\.{0,2}( +|$)", expr))
    parsed = expr[:matches[0].start()]
    for match in matches:
        left, content, 
        vector = f"vec({match.group(2)})"
        parsed += expr.replace(match.string, match.group(1) + vector + match.group(4))
        if matches.peek(None):
            parsed += expr[match.end():matches.peek().start()]
        else:
            # we're on the last match
            parsed += expr[match.end():]
    return parsed

def parse_dot_derivatives(expr: str) -> str:
    """
    Transforms the shorthand dot derivatives notation into a standard command call

    >>> parse_dot_derivatives('OM->.. = (r.. - r theta.^2)__(a_r) u_r-> + (2 r. theta. + r theta..)__(a_theta) u_theta-> + z.. u_z->')
    'ddot(OM->) = (ddot(r) - r dot(theta)^2)__(a_r) u_r-> + (2 dot(r) dot(theta) + r ddot(theta))__(a_theta) u_theta-> + ddot(z) u_z->'
    """
    matches = peekable(re.finditer(r"(^| +|\()([^.\( ]+)(\.{1,2})( +|\)|\^|_|$)", expr))
    parsed = expr[:matches[0].start()]
    for match in matches:
        spacing_left, dotted_expr, dots, spacing_right = match.groups()
        print(match, '\n', repr(f'{spacing_left}dots{{{dotted_expr}, {len(dots)}}}{spacing_right}'))
        parsed += expr.replace(match.string, spacing_left + ('d' * len(dots) + 'ot') + '(' + dotted_expr + ')' + spacing_right)
        if matches.peek(None):
            parsed += expr[match.end():matches.peek().start()]
        else:
            # we're on the last match
            parsed += expr[match.end():]
    return parsed

def parse_goes_to_arrow(expr: str) -> str:
    """
    Transforms the "goes to" arrow used to express limits into a standard arrow with underscript

    >>> parse_goes_to_arrow('a --(n -> oo)-> 5 and b --+oo-> 6')
    'a ->__(n -> oo) 5 and b ->__(+oo) 6'
    """
    matches = peekable(r"(\s+)--(\S+|\(.+\))->(\s+)")
    parsed = expr[:matches[0].start()]
    for match in matches:
        spacing_left, content, spacing_right = match.groups()
        parsed += expr.replace(match.string, f'{spacing_left}->__{content}{spacing_right}')
        if matches.peek(None):
            parsed += expr[matches.end():matches.peek().start()]
        else:
            parsed += expr[match.end():]
    return parsed


def get_font_name(char: str) -> str:
    return {
        "op": "operatorname",
        "bb": "mathbb",
        "c": "mathcal",
        "b": "mathbold",
        "f": "mathcur",
        "k": "mathfrak",
    }.get(char, "mathnorm")


preamble, document = split_preamble_document(source)
preamble, macros = parse_preamble(preamble)
